#
# Copyright (c) 2011, 2013, 2014, 2015, 2016
#  Adam Tauno Williams <awilliam@whitemice.org>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
from coils.core import ServerDefaultsManager, BLOBManager
from coils.net import PathObject
from coils.net.ossf import MarshallOSSFChain


class DocumentObject(PathObject):
    _MIME_MAP_ = None

    def __init__(self, parent, name, **params):
        self.entity = None
        self.request = None
        self.parameters = None
        self.name = name
        self.disposition = 'attachment'
        PathObject.__init__(self, parent, **params)
        if DocumentObject._MIME_MAP_ is None:
            sd = ServerDefaultsManager()
            self._mime_type_map = sd.default_as_dict('CoilsExtensionMIMEMap')

    #
    # Method Handlers
    #

    def object_for_key(self, name, auto_load_enabled=True, is_webdav=False):
        if (name == 'download'):
            self.disposition = 'attachment'
        elif (name == 'view'):
            self.disposition = 'inline'
        return self

    def get_mime_type(self):
        '''
        Return the MIME type of the document.  Previously this method did
        some default inspection and other monkeying around; this process has
        all been unified into the TypeManager's get_mimetype method - so this
        method is now just a proxy for the TM's method.
        '''
        return self.context.type_manager.get_mimetype(self.entity)

    def do_HEAD(self):
        # TODO: Check for locks!

        headers = {
            'etag': '{0}:{1}'.format(
                self.entity.object_id,
                self.entity.version,
            ),
            'X-OpenGroupware-Filename': self.entity.get_display_name(),
            'X-OpenGroupware-Content-Length': str(self.entity.file_size),
            'X-OpenGroupware-Document-Id': str(self.entity.object_id),
            'X-OpenGroupware-Folder-Id': str(self.entity.folder_id),
            'X-OpenGroupware-Revision': str(self.entity.version_count),
            'X-OpenGroupware-Version': str(self.entity.version),
            'X-OpenGroupware-Project-Id': str(self.entity.project_id),
            'X-OpenGroupware-Folder-Path': '/'.join(
                self.entity.folder_name_path
            ),
        }

        if self.entity.versions:
             headers['X-OpenGroupware-Checksum'] = str(self.entity.checksum)

        self.request.simple_response(
            200,
            data=None,
            mimetype=self.get_mime_type(),
            headers=headers,
        )

    def do_GET(self):
        # TODO: Check for locks!
        handle = self.context.run_command(
            'document::get-handle',
            id=self.entity.object_id,
        )
        mimetype = self.get_mime_type()
        handle, mimetype = \
            MarshallOSSFChain(handle, mimetype, self.parameters)
        self.log.debug(
            'MIME-Type after OSSF processing is {0}'.format(mimetype))
        self.context.run_command(
            'document::record-download',
            document=self.entity,
        )
        self.context.commit()

        headers = {
            'Content-Disposition': '{0}; filename={1}'.format(
                self.disposition,
                self.entity.get_display_name(),
            ),
            'X-OpenGroupware-Filename': self.entity.get_display_name(),
            'etag': '{0}:{1}'.format(
                self.entity.object_id,
                self.entity.version,
            ),
            'X-OpenGroupware-Content-Length': str(self.entity.file_size),
            'X-OpenGroupware-Document-Id': str(self.entity.object_id),
            'X-OpenGroupware-Folder-Id': str(self.entity.folder_id),
            'X-OpenGroupware-Revision': str(self.entity.version_count),
            'X-OpenGroupware-Version': str(self.entity.version),

            'X-OpenGroupware-Project-Id': str(self.entity.project_id),
            'X-OpenGroupware-Folder-Path': '/'.join(
                [
                    str(x) if x is not None else ''
                    for x in self.entity.folder_name_path
                ]
            ),
        }

        if self.entity.versions:
            headers['X-OpenGroupware-Checksum'] = str(self.entity.checksum)

        self.request.stream_response(
            200,
            stream=handle,
            mimetype=mimetype,
            headers=headers,
        )

        BLOBManager.Close(handle)
